# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Experimental ops"."""

from typing import Tuple

from jax._src.lib.mlir import ir
from jax.interpreters import mlir
import jax.numpy as jnp

from tf2jax._src import ops
from tf2jax._src.experimental import mhlo


@ops.register_operation("XlaCallModule")
def _xla_call_module(proto):
  """Parse a XlaCallModule op."""
  ops._check_attrs(  # pylint: disable=protected-access
      proto,
      {"Tin", "Sout", "Tout", "dim_args_spec", "module", "version"})

  version = proto.attr["version"].i
  if version != 2:
    raise ValueError(
        f"Only version 2 (StableHLO) is supported, found {version}.")

  dim_args_spec = tuple(proto.attr["dim_args_spec"].list.s)
  if dim_args_spec:
    raise ValueError("Dynamic shapes is not yet supported, found "
                     f"dim_args_spec={dim_args_spec}.")

  with mlir.make_ir_context():
    module = ir.Module.parse(proto.attr["module"].s)
  mhlo_text = mlir.module_to_string(module)

  def _func(*operands: jnp.ndarray) -> Tuple[jnp.ndarray, ...]:
    return mhlo.mhlo_apply(*operands, mhlo_text=mhlo_text)

  return _func



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tf2jax."""

from absl.testing import absltest
from absl.testing import parameterized

import jax.numpy as jnp
import numpy as np
import tensorflow as tf
from tf2jax._src import numpy_compat

_dtypes = [
    tf.bool, tf.uint8, tf.uint16, tf.uint32, tf.uint64, tf.int8, tf.int16,
    tf.int32, tf.int64, tf.bfloat16, tf.float16, tf.float32, tf.float64,
    tf.complex64, tf.complex128
]


class NumpyCompatTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ("np", np, numpy_compat.tf_to_np_dtypes),
      ("jnp", jnp, numpy_compat.tf_to_jnp_dtypes),
  )
  def test_dtype_conversion(self, np_module, dtype_map):
    self.assertEqual(len(_dtypes), len(dtype_map))
    for src in _dtypes:
      dst = "bool_" if src.name == "bool" else src.name
      if src.name == "bfloat16":
        self.assertIs(dtype_map[src], jnp.bfloat16)
      else:
        self.assertIs(dtype_map[src], getattr(np_module, dst))


if __name__ == "__main__":
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests tf2jax."""

import inspect

from absl.testing import parameterized

import chex
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

import sonnet as snt
import tensorflow as tf
from tf2jax._src import config
from tf2jax._src import tf2jax
import tree


class TestDense(tf.Module):

  def __init__(self, input_dim, output_size, name=None):
    super().__init__(name=name)
    self.w = tf.Variable(tf.random.normal([input_dim, output_size]), name="w")
    self.b = tf.Variable(tf.zeros([output_size]), name="b")

  def __call__(self, x):
    y = tf.matmul(x, self.w) + self.b
    return tf.nn.relu(y)


class TestMLP(tf.Module):
  """Variables for different dense layers will not have unique names."""

  def __init__(self, input_size, sizes, name=None):
    super().__init__(name=name)
    self.layers = []
    with self.name_scope:
      for size in sizes:
        self.layers.append(TestDense(input_dim=input_size, output_size=size))
        input_size = size

  @tf.Module.with_name_scope
  def __call__(self, x):
    for layer in self.layers:
      x = layer(x)
    return x


class ModelsTest(tf.test.TestCase, parameterized.TestCase):

  def _test_convert(self, tf_func, inputs):
    jax_func, jax_params = tf2jax.convert(
        tf.function(tf_func), np.zeros_like(inputs))
    jax_func = self.variant(jax_func)

    jax_results, jax_params = jax_func(jax_params, inputs)
    tf_results, tf_params = tf_func(inputs)

    # Check outputs
    for tf_res, jax_res in zip(
        tree.flatten(tf_results), tree.flatten(jax_results)):
      self.assertAllClose(tf_res.numpy(), jax_res, atol=1e-4)

    # Check params (batchnorm stats changed).
    for tf_var in tree.flatten(tf_params):
      jax_var = jax_params[tf_var.name.split(":")[0]]
      self.assertAllClose(tf_var.numpy(), jax_var, atol=1e-5)

  @chex.variants(with_jit=True, without_jit=True)
  def test_mlp(self):
    tf.random.set_seed(42)

    inputs = np.linspace(-1., 1., 128 * 16, dtype=np.float32).reshape((128, 16))

    model = snt.nets.MLP((64, 10,))
    def tf_func(x):
      outputs = model(x)
      return outputs, model.variables

    self._test_convert(tf_func, inputs)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("inference", False), ("training", True)),
          named=True,
      ))
  def test_resnet(self, training):
    tf.random.set_seed(42)

    inputs = np.linspace(
        -1., 1., 2 * 64 * 64 * 3, dtype=np.float32).reshape((2, 64, 64, 3))

    model = snt.nets.ResNet([1, 1, 1, 1], 10)
    def tf_func(x):
      outputs = model(x, is_training=training)
      return outputs, model.variables

    self._test_convert(tf_func, inputs)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("inference", False), ("training", True)),
          named=True,
      ))
  def test_vqvae(self, training):
    tf.random.set_seed(42)

    inputs = np.linspace(
        -1., 1., 2 * 64 * 64 * 3, dtype=np.float32).reshape((2, 64, 64, 3))

    model = snt.nets.VectorQuantizer(3, 100, 0.1)
    def tf_func(x):
      return model(x, is_training=training), model.variables

    self._test_convert(tf_func, inputs)


class FeaturesTest(tf.test.TestCase, parameterized.TestCase):

  def _setup_saved_model(self, *inputs):
    def tf_func(x):
      return tf.exp(x)

    # Save.
    model = tf.Module()
    model.f = tf.function(tf_func)
    for inp in inputs:
      if isinstance(inp, tf.TensorSpec):
        model.f.get_concrete_function(inp)  # Dummy call.
      else:
        model.f(np.zeros_like(inp))  # Dummy call.
    tmp_dir = self.create_tempdir()
    tf.saved_model.save(model, tmp_dir.full_path)

    return tf_func, tf.saved_model.load(tmp_dir.full_path)

  @chex.variants(with_jit=True, without_jit=True)
  def test_saved_model(self):
    dummy_inputs = np.zeros([10, 5], dtype=np.float32)
    tf_func, restored = self._setup_saved_model(dummy_inputs)

    # Convert to Jax (with and without explicit inputs).
    for inp in [(), (dummy_inputs,)]:
      if inp:
        jax_func, _ = tf2jax.convert(restored.f, *inp)
      else:
        jax_func, _ = tf2jax.convert_from_restored(restored.f)
      jax_func = self.variant(jax_func)

      test_inputs = np.ones([20, 5], dtype=np.float32)
      expected_outputs = tf_func(test_inputs)
      with config.override_config("strict_shape_check", False):
        actual_outputs, _ = jax_func({}, test_inputs)
      self.assertAllClose(expected_outputs, actual_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  def test_saved_model_ambiguous(self):
    dummy_one = np.zeros([10, 5], dtype=np.float32)
    dummy_two = np.zeros([10, 5, 3], dtype=np.float32)
    tf_func, restored = self._setup_saved_model(dummy_one, dummy_two)

    with self.assertRaisesRegex(ValueError, "Found 2 concrete functions"):
      jax_func, _ = tf2jax.convert_from_restored(restored.f)
    jax_func, _ = tf2jax.convert(restored.f, dummy_two)
    jax_func = self.variant(jax_func)

    test_inputs = np.ones([20, 7, 2], dtype=np.float32)
    expected_outputs = tf_func(test_inputs)
    with config.override_config("strict_shape_check", False):
      actual_outputs, _ = jax_func({}, test_inputs)
    self.assertAllClose(expected_outputs, actual_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  def test_saved_model_functional(self):
    dummy_inputs = np.zeros([10, 5], dtype=np.float32)
    tf_func, restored = self._setup_saved_model(dummy_inputs)

    jax_func = tf2jax.convert_functional_from_restored(restored.f)
    jax_func = self.variant(jax_func)

    test_inputs = np.ones([20, 5], dtype=np.float32)
    expected_outputs = tf_func(test_inputs)
    with config.override_config("strict_shape_check", False):
      actual_outputs = jax_func(test_inputs)
    self.assertAllClose(expected_outputs, actual_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  def test_saved_model_polymorphic_shape(self):
    dummy_input_spec = tf.TensorSpec((None, None, 2), dtype=tf.float32)
    tf_func, restored = self._setup_saved_model(dummy_input_spec)

    jax_func = tf2jax.convert_functional_from_restored(restored.f)
    jax_func = self.variant(jax_func)

    test_inputs = np.ones([10, 5, 2], dtype=np.float32)
    expected_outputs = tf_func(test_inputs)
    actual_outputs = jax_func(test_inputs)
    self.assertAllClose(expected_outputs, actual_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  def test_strict_check(self):
    def tf_func(x):
      return x.shape

    orig_inputs = np.zeros((10, 5), dtype=np.float32)

    model = tf.Module()
    model.f = tf.function(tf_func)
    model.f(orig_inputs)
    tmp_dir = self.create_tempdir()
    tf.saved_model.save(model, tmp_dir.full_path)
    restored = tf.saved_model.load(tmp_dir.full_path)

    jax_func = tf2jax.convert_functional_from_restored(restored.f)
    jax_func = self.variant(jax_func)

    with self.subTest("valid_input_shape"):
      expected_outputs = tf_func(orig_inputs)
      with config.override_config("strict_shape_check", True):
        actual_outputs = jax_func(orig_inputs)
        self.assertAllClose(expected_outputs, actual_outputs)
      with config.override_config("strict_dtype_check", True):
        actual_outputs = jax_func(orig_inputs)
        self.assertAllClose(expected_outputs, actual_outputs)

    test_inputs = np.ones([10, 5, 2], dtype=np.float32)
    with self.subTest("invalid_input_shape"):
      expected_outputs = tf_func(test_inputs)
      with config.override_config("strict_shape_check", True):
        with self.assertRaisesRegex(ValueError, "incompatible input shape"):
          actual_outputs = jax_func(test_inputs)
      with config.override_config("strict_shape_check", False):
        actual_outputs = jax_func(test_inputs)
        self.assertNotAllClose(expected_outputs, actual_outputs)

    test_inputs = np.ones([10, 5], dtype=np.int32)
    with self.subTest("invalid_input_dtype"):
      expected_outputs = tf_func(test_inputs)
      with config.override_config("strict_dtype_check", True):
        with self.assertRaisesRegex(ValueError, "incompatible input dtype"):
          actual_outputs = jax_func(test_inputs)
      with config.override_config("strict_dtype_check", False):
        actual_outputs = jax_func(test_inputs)
        self.assertAllClose(expected_outputs, actual_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("float32", tf.float32), ("float64", tf.float64)),
          named=True,
      ))
  def test_force_bf16_consts(self, tf_dtype):
    @tf.function
    def tf_func(x):
      return x + tf.constant(3.14, dtype=x.dtype)

    np_inputs = np.array(42.0, tf_dtype.as_numpy_dtype())
    tf_outputs = tf_func(np_inputs)

    dtype_config_name = ("force_const_float32_to_bfloat16"
                         if tf_dtype == tf.float32 else
                         "force_const_float64_to_bfloat16")

    # This is the default.
    with config.override_config(dtype_config_name, False):
      jax_func = tf2jax.convert_functional(tf_func, np_inputs)
    jax_func = self.variant(jax_func)
    orig_jax_outputs = jax_func(np_inputs)
    self.assertEqual(
        jnp.array(np_inputs).dtype,
        jnp.array(orig_jax_outputs).dtype)
    self.assertAllClose(tf_outputs, orig_jax_outputs)

    with config.override_config(dtype_config_name, True):
      jax_func = tf2jax.convert_functional(tf_func, np_inputs)
    jax_func = self.variant(jax_func)
    forced_jax_outputs = jax_func(jnp.asarray(np_inputs, jnp.bfloat16))
    self.assertEqual(jnp.bfloat16, forced_jax_outputs.dtype)
    self.assertAllClose(tf.cast(tf_outputs, tf.bfloat16), forced_jax_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  def test_force_bf16_consts_for_leaks(self):
    @tf.function
    def tf_func(x):
      return x + tf.constant(3.14, dtype=tf.float32)

    np_inputs = np.array(42.0, np.float32)
    tf_outputs = tf_func(np_inputs)

    class CachedFn(hk.Module):

      cache = []

      def __init__(self):
        super().__init__(name=None)
        if not self.cache:
          with config.override_config("force_const_float32_to_bfloat16", True):
            self.cache.append(jax.jit(
                tf2jax.convert_functional(tf_func, np_inputs)))

      def __call__(self, x):
        return self.cache[0](x)

    def call_cached_fn(x):
      return CachedFn()(x)

    cached_fn = hk.without_apply_rng(hk.transform(call_cached_fn))
    params = self.variant(cached_fn.init)(jax.random.PRNGKey(42), np_inputs)
    jax_inputs = jnp.asarray(np_inputs, jnp.bfloat16)
    jax_outputs = self.variant(cached_fn.apply)(params, jax_inputs)

    self.assertEqual(jnp.bfloat16, jax_outputs.dtype)
    self.assertAllClose(tf.cast(tf_outputs, tf.bfloat16), jax_outputs)

  def test_static_argnums(self):
    @tf.function
    def tf_func(x):
      return tf.zeros((x,))

    with self.subTest("literal_input"):
      jax_func = tf2jax.convert_functional(tf_func, 10)
      self.assertEqual(jax_func(10).shape, (10,))
      with self.assertRaisesRegex(ValueError, "Found unexpected literal value"):
        jax_func(13)
      with self.assertRaisesRegex(ValueError, "Found unexpected tracer"):
        jax.jit(jax_func)(13)
      with self.assertRaisesRegex(ValueError, "Found unexpected literal value"):
        jax.jit(jax_func, static_argnums=0)(13)

    with self.subTest("array_input"):
      jax_func = tf2jax.convert_functional(tf_func, np.array(10))
      self.assertEqual(jax_func(10).shape, (10,))
      self.assertEqual(jax_func(13).shape, (13,))
      with self.assertRaisesRegex(
          TypeError,
          "Shapes must be 1D sequences of concrete values of integer type"):
        jax.jit(jax_func)(13)
      self.assertEqual(jax.jit(jax_func, static_argnums=0)(13).shape, (13,))

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("with_custom_gradient", True), ("without_custom_gradient", False)),
          named=True,
      ))
  def test_custom_gradient(self, use_custom_gradient):
    @tf.function
    @tf.custom_gradient
    def tf_func(x):
      e = tf.exp(x)
      def grad(dy):
        # This is deliberately the wrong gradient.
        return dy * (1 - 1 / (1 + e)) * tf.sin(x) + 0.42
      return tf.reduce_sum(tf.math.log(1 + e)), grad

    np_inputs = np.array(range(6), dtype=np.float32).reshape(3, 2)
    tf_inputs = tf.constant(np_inputs)
    with tf.GradientTape() as tape:
      tape.watch(tf_inputs)
      tf_outputs = tf_func(tf_inputs)
      tf_grads = tape.gradient(tf_outputs, tf_inputs)

    with config.override_config("convert_custom_gradient", use_custom_gradient):
      jax_func = tf2jax.convert_functional(tf_func, np.zeros_like(np_inputs))
    jax_func = self.variant(jax_func)
    jax_outputs = jax_func(np_inputs)
    jax_grads = jax.grad(jax_func)(np_inputs)

    self.assertAllClose(tf_outputs, jax_outputs)
    if use_custom_gradient:
      self.assertAllClose(tf_grads, jax_grads)
    else:
      self.assertNotAllClose(tf_grads, jax_grads)

  @chex.variants(with_jit=True, without_jit=True)
  def test_trainable(self):
    can_train = tf.Variable(3.14, trainable=True, name="can_train")
    not_train = tf.Variable(42., trainable=False, name="not_train")

    @tf.function
    def tf_func(x):
      return x + can_train + not_train

    np_inputs = np.array(range(6), dtype=np.float32).reshape(3, 2)
    jax_func, jax_params = tf2jax.convert(tf_func, np.zeros_like(np_inputs))
    self.assertTrue(jax_params["can_train"].trainable)
    self.assertFalse(jax_params["not_train"].trainable)

    tf_outputs = tf_func(tf.constant(np_inputs))
    jax_outputs, _ = self.variant(jax_func)(jax_params, np_inputs)
    self.assertAllClose(tf_outputs, jax_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (
              ("positional", ((3.14, 42.0), {})),
              ("positional_and_keyword", ((3.14,), dict(y=42.0))),
              ("keyword", ((), dict(x=3.14, y=42.0))),
          ),
          named=True,
      ))
  def test_positional_and_keyword_args(self, all_args):
    @tf.function
    def tf_func(x, y):
      return x + y

    fn_args, fn_kwargs = tree.map_structure(np.array, all_args)
    tf_outputs = tf_func(*fn_args, **fn_kwargs)

    zero_args, zero_kwargs = tree.map_structure(np.zeros_like, all_args)
    jax_func = tf2jax.convert_functional(tf_func, *zero_args, **zero_kwargs)
    jax_outputs = self.variant(jax_func)(*fn_args, **fn_kwargs)

    self.assertAllClose(tf_outputs, jax_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (
              ("no_defaults", lambda *, kw0, kw1: kw0 + kw1),
              ("some_defaults", lambda *, kw0=3.14, kw1: kw0 + kw1),
              ("all_defaults", lambda *, kw0=3.14, kw1=42.0: kw0 + kw1),
              ("all_defaults_some_used",
               lambda *, kw0=3.14, kw1=42.0, kw2=2.8: kw0 + kw1 + kw2),
          ),
          named=True,
      ))
  def test_keyword_only(self, fn):
    tf_func = tf.function(fn)

    inputs = (np.array(3.5, dtype=np.float32), np.array(7.2, dtype=np.float32))
    tf_outputs = tf_func(kw0=inputs[0], kw1=inputs[1])

    jax_func = tf2jax.convert_functional(
        tf_func, kw0=np.zeros_like(inputs[0]), kw1=np.zeros_like(inputs[1]))
    jax_outputs = self.variant(jax_func)(kw0=inputs[0], kw1=inputs[1])

    self.assertAllClose(tf_outputs, jax_outputs)

  def test_metadata(self):
    @tf.function
    def tf_func(x, y, *, z):
      return x + y + z

    x = np.zeros((10, 5), np.float32)
    y = np.zeros((10, 1), np.float32)
    z = np.zeros((), np.float32)
    jax_func = tf2jax.convert_functional(tf_func, x, y=y, z=z)
    self.assertEqual(jax_func.structured_inputs, (
        (
            tf.TensorSpec(shape=(10, 5), dtype=tf.float32, name="x"),
            tf.TensorSpec(shape=(10, 1), dtype=tf.float32, name="y"),
        ),
        {
            "z": tf.TensorSpec(shape=(), dtype=tf.float32, name="z")
        },
    ))
    self.assertEqual(
        jax_func.structured_outputs,
        tf.TensorSpec(shape=(10, 5), dtype=tf.float32, name="Identity"))

    expected_sig = inspect.Signature([
        inspect.Parameter("x", inspect.Parameter.POSITIONAL_OR_KEYWORD),
        inspect.Parameter("y", inspect.Parameter.POSITIONAL_OR_KEYWORD),
        inspect.Parameter("z", inspect.Parameter.KEYWORD_ONLY),
    ])
    self.assertEqual(expected_sig, jax_func.signature)
    self.assertEqual(expected_sig, inspect.signature(jax_func))

  @chex.variants(with_jit=True, without_jit=True)
  def test_jit_nesting(self):
    @tf.function
    def tf_fn(x):
      return tf.cos(x)

    @jax.jit
    def tf2jax_fn(x):
      jax_fn = tf2jax.convert_functional(tf_fn, x)
      return jnp.sin(self.variant(jax_fn)(x))

    inputs = np.linspace(-1., 1., 6, dtype=np.float32).reshape((2, 3))
    self.assertAllClose(tf.sin(tf_fn(inputs)), tf2jax_fn(inputs))

  @chex.variants(with_jit=True, without_jit=True)
  def test_non_unique_variable_names(self):
    model = TestMLP(input_size=5, sizes=[7, 8])
    self.assertNotEqual(
        len(set([v.name for v in model.variables])), len(model.variables))

    @tf.function
    def forward(x):
      return model(x)

    x = np.arange(10).reshape(2, 5).astype(np.float32)
    tf_outputs = forward(x)

    apply_fn, params = tf2jax.convert(forward, tf.TensorSpec((None, 5)))
    jax_outputs, _ = self.variant(apply_fn)(params, x)

    self.assertAllClose(tf_outputs, jax_outputs)


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Util functions."""

import inspect


def fullargspec_to_signature(fullargspec) -> inspect.Signature:
  """Convert a {inspect|tf}.FullArgSpec to a inspect.Signature."""
  default_offset = len(fullargspec.args) - len(fullargspec.defaults or ())
  parameters = []
  # Positional or keyword args.
  for idx, arg in enumerate(fullargspec.args):
    param_dict = dict(name=arg, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
    if idx >= default_offset:
      param_dict["default"] = fullargspec.defaults[idx - default_offset]
    parameters.append(inspect.Parameter(**param_dict))
  # Varargs
  if fullargspec.varargs is not None:
    parameters.append(
        inspect.Parameter(
            fullargspec.varargs, kind=inspect.Parameter.VAR_POSITIONAL))
  # Keyword-only args.
  for arg in fullargspec.kwonlyargs:
    param_dict = dict(name=arg, kind=inspect.Parameter.KEYWORD_ONLY)
    if arg in (fullargspec.kwonlydefaults or {}):
      param_dict["default"] = fullargspec.kwonlydefaults[arg]
    parameters.append(inspect.Parameter(**param_dict))
  # Kwargs.
  if fullargspec.varkw is not None:
    parameters.append(
        inspect.Parameter(
            fullargspec.varkw, kind=inspect.Parameter.VAR_KEYWORD))
  signature = inspect.Signature(parameters)

  return signature



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MHLO JAX primitive"."""

import functools
import threading
from typing import Any, Dict, Tuple

import jax
from jax import core
from jax._src.lib import xla_client as xc
import jax._src.util as util  # pylint: disable=consider-using-from-import
from jax.interpreters import mlir

from jaxlib.mlir import ir
from jaxlib.mlir.dialects import mhlo

import numpy as np

_program_cache: Dict[str, Tuple[str, xc._xla.XlaComputation]] = {}
_program_lock = threading.Lock()

mhlo_apply_p = core.Primitive("mhlo_apply")
mhlo_apply_p.multiple_results = True


def _get_program(mhlo_text: str) -> Tuple[str, xc._xla.XlaComputation]:
  """Thread-safe cache of MHLO string to XlaComputation."""
  try:
    return _program_cache[mhlo_text]
  except KeyError:
    pass

  with _program_lock:
    program = xc._xla.mlir.mlir_module_to_xla_computation(  # pylint: disable=protected-access
        mlir_module=mhlo_text,
        use_tuple_args=False,
        return_tuple=True)
    named_program = (f"_tf2jax_user_mhlo_fn_{len(_program_cache):06}", program)
    _program_cache[mhlo_text] = named_program
  return named_program


@functools.lru_cache(None)
def _get_compiled_program(mhlo_text: str, backend: Any) -> Any:
  executable = backend.compile(mhlo_text)
  return executable


def mhlo_apply(*args, mhlo_text: str):
  ret = mhlo_apply_p.bind(*args, mhlo_text=mhlo_text)
  # TODO(shaobohou) Unpack singleton result?
  if len(ret) == 1:
    return ret[0]
  else:
    return tuple(ret)


def mhlo_apply_impl(*args, mhlo_text: str):
  backend = jax.lib.xla_bridge.get_backend()
  executable = _get_compiled_program(mhlo_text, backend)
  return jax.lib.xla_client.execute_with_python_values(
      executable=executable,
      arguments=[np.asarray(x) for x in args],
      backend=backend)

mhlo_apply_p.def_impl(mhlo_apply_impl)


def mhlo_apply_abstract_eval(*args, mhlo_text: str):
  del args
  _, program = _get_program(mhlo_text)
  result_spec = program.program_shape().result_shape()
  assert result_spec.is_tuple()
  return tuple([
      jax.ShapedArray(spec.dimensions(), spec.element_type())
      for spec in result_spec.tuple_shapes()
  ])

mhlo_apply_p.def_abstract_eval(mhlo_apply_abstract_eval)


def mhlo_apply_lowering(ctx: mlir.LoweringRuleContext, *args, mhlo_text: str):
  """Lowering rule."""
  program_name, program = _get_program(mhlo_text)
  assert program.program_shape().result_shape().is_tuple()

  module_ctx = ctx.module_context
  if xc.mlir_api_version < 41:
    mhlo_module = mlir.xla_computation_to_mhlo_module(program)
    callee_name = mlir.merge_mhlo_modules(
        dst_module=module_ctx.module,
        sym_name=program_name,
        src_module=mhlo_module)
  else:
    mhlo_module = mlir.xla_computation_to_mlir_module(program)
    callee_name = mlir.merge_mlir_modules(
        dst_module=module_ctx.module,
        sym_name=program_name,
        src_module=mhlo_module)

  output_types = tuple(map(mlir.aval_to_ir_types, ctx.avals_out))
  flat_output_types = util.flatten(output_types)
  output_type = ir.TupleType.get_tuple(flat_output_types)

  call = mlir.func_dialect.CallOp([output_type],
                                  ir.FlatSymbolRefAttr.get(callee_name),
                                  mlir.flatten_lowering_ir_args(args)).result
  flat_results = [mhlo.GetTupleElementOp(call, mlir.i32_attr(i)).result
                  for i in range(len(flat_output_types))]

  return tuple(util.unflatten(flat_results, list(map(len, output_types))))

mlir.register_lowering(mhlo_apply_p, mhlo_apply_lowering)



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MHLO primitive tests."""

import functools

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized

import chex
import jax
import numpy as np

from tf2jax._src.experimental import mhlo


def _convert_to_mhlo(jax_fn, inputs, *, dialect):
  lowered_forward = jax_fn.lower(*inputs)
  mhlo_text = lowered_forward.as_text(dialect=dialect)
  return mhlo_text


def _check_transforms(fn, inputs, *, dialect):
  jaxpr = jax.make_jaxpr(fn)(*inputs)
  logging.info(jaxpr)

  mhlo_text = jax.jit(fn).lower(*inputs).as_text(dialect=dialect)
  logging.info(mhlo_text)


class MhloTest(chex.TestCase):

  def _assert_all_close(self, expect_fn, actual_fn, inputs):
    expect_outputs = self.variant(expect_fn)(*inputs)
    actual_outputs = self.variant(actual_fn)(*inputs)
    logging.info(expect_outputs)
    logging.info(actual_outputs)
    chex.assert_trees_all_close(expect_outputs, actual_outputs)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      ("mhlo", "mhlo"),
      ("stablehlo", "stablehlo"),
  )
  def test_one_input_and_one_output(self, dialect):
    @jax.jit
    def fn(x):
      return x * 2 + 3.14

    inputs = (np.ones((3, 2), dtype=np.float32) * 10,)
    mhlo_text = _convert_to_mhlo(
        fn, jax.tree_map(np.zeros_like, inputs), dialect=dialect)
    chex.assert_trees_all_close(
        mhlo.mhlo_apply(*inputs, mhlo_text=mhlo_text), fn(*inputs))

    def make_top_fn(sub_fn):
      def top_fn(x):
        return sub_fn(x + 8) * 10
      return top_fn

    expect_top_fn = make_top_fn(fn)
    actual_top_fn = make_top_fn(
        functools.partial(mhlo.mhlo_apply, mhlo_text=mhlo_text))
    self._assert_all_close(expect_top_fn, actual_top_fn, inputs)
    _check_transforms(actual_top_fn, inputs, dialect=dialect)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      ("mhlo", "mhlo"),
      ("stablehlo", "stablehlo"),
  )
  def test_two_inputs_and_one_output(self, dialect):
    @jax.jit
    def fn(x, y):
      return x * 2 + y

    inputs = (
        np.ones((3, 2), dtype=np.float32) * 10,
        np.ones((3, 2), dtype=np.float32) * 20,
    )
    mhlo_text = _convert_to_mhlo(
        fn, jax.tree_map(np.zeros_like, inputs), dialect=dialect)
    chex.assert_trees_all_close(
        mhlo.mhlo_apply(*inputs, mhlo_text=mhlo_text), fn(*inputs))

    def make_top_fn(sub_fn):
      def top_fn(x, y):
        return sub_fn(x + 8, y + 9) * 10
      return top_fn

    expect_top_fn = make_top_fn(fn)
    actual_top_fn = make_top_fn(
        functools.partial(mhlo.mhlo_apply, mhlo_text=mhlo_text))
    self._assert_all_close(expect_top_fn, actual_top_fn, inputs)
    _check_transforms(actual_top_fn, inputs, dialect=dialect)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      ("mhlo", "mhlo"),
      ("stablehlo", "stablehlo"),
  )
  def test_two_inputs_and_two_outputs(self, dialect):
    @jax.jit
    def fn(x, y):
      return x * 2 + y, x + y * 3

    inputs = (
        np.ones((3, 2), dtype=np.float32) * 10,
        np.ones((3, 2), dtype=np.float32) * 20,
    )
    mhlo_text = _convert_to_mhlo(
        fn, jax.tree_map(np.zeros_like, inputs), dialect=dialect)
    chex.assert_trees_all_close(
        mhlo.mhlo_apply(*inputs, mhlo_text=mhlo_text), fn(*inputs))

    def make_top_fn(sub_fn):
      def top_fn(x, y):
        res0, res1 = sub_fn(x + 8, y + 9)
        return res0 * 10, res1 * 3.14
      return top_fn

    expect_top_fn = make_top_fn(fn)
    actual_top_fn = make_top_fn(
        functools.partial(mhlo.mhlo_apply, mhlo_text=mhlo_text))
    self._assert_all_close(expect_top_fn, actual_top_fn, inputs)
    _check_transforms(actual_top_fn, inputs, dialect=dialect)


if __name__ == "__main__":
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utililties for wrapping around numpy and jax.numpy."""

import itertools

from typing import Sequence, Union

import jax
import jax.numpy as jnp
import numpy as np
import tensorflow as tf


tf_to_np_dtypes = {
    tf.bool: np.bool_,
    tf.uint8: np.uint8,
    tf.uint16: np.uint16,
    tf.uint32: np.uint32,
    tf.uint64: np.uint64,
    tf.int8: np.int8,
    tf.int16: np.int16,
    tf.int32: np.int32,
    tf.int64: np.int64,
    tf.bfloat16: jnp.bfloat16,
    tf.float16: np.float16,
    tf.float32: np.float32,
    tf.float64: np.float64,
    tf.complex64: np.complex64,
    tf.complex128: np.complex128,
}


tf_to_jnp_dtypes = {
    tf.bool: jnp.bool_,
    tf.uint8: jnp.uint8,
    tf.uint16: jnp.uint16,
    tf.uint32: jnp.uint32,
    tf.uint64: jnp.uint64,
    tf.int8: jnp.int8,
    tf.int16: jnp.int16,
    tf.int32: jnp.int32,
    tf.int64: jnp.int64,
    tf.bfloat16: jnp.bfloat16,
    tf.float16: jnp.float16,
    tf.float32: jnp.float32,
    tf.float64: jnp.float64,
    tf.complex64: jnp.complex64,
    tf.complex128: jnp.complex128,
}


_NP_LIKES = (np.ndarray, np.number, np.bool_, bool, int, float)


# We avoids importing jax2tf.shape_poly.is_poly_dim as jax2tf often depends
# on the most recent tensorflow version.
def _is_poly_dim(x):
  cls = type(x)
  return (cls.__module__ == "jax.experimental.jax2tf.shape_poly" and
          cls.__name__ == "_DimPolynomial")


def _get_np(*args):
  """Select numpy backend based on input types."""
  no_jax = all((isinstance(x, _NP_LIKES) or _is_poly_dim(x)) for x in args)
  return np if no_jax else jnp


def _get_dtypes(*args):
  """Select dtypes backend based on input types."""
  no_jax = _get_np(*args) is np
  return tf_to_np_dtypes if no_jax else tf_to_jnp_dtypes


def get_jax_dtype(dtype: tf.DType):
  return tf_to_jnp_dtypes[dtype]


# Simple numeric ops.
add = lambda x, y: _get_np(x, y).add(x, y)
equal = lambda x, y: _get_np(x, y).equal(x, y)
floor_divide = lambda x, y: _get_np(x, y).floor_divide(x, y)
greater = lambda x, y: _get_np(x, y).greater(x, y)
greater_equal = lambda x, y: _get_np(x, y).greater_equal(x, y)
less = lambda x, y: _get_np(x, y).less(x, y)
less_equal = lambda x, y: _get_np(x, y).less_equal(x, y)
maximum = lambda x, y: _get_np(x, y).maximum(x, y)
minimum = lambda x, y: _get_np(x, y).minimum(x, y)
mod = lambda x, y: _get_np(x, y).mod(x, y)
multiply = lambda x, y: _get_np(x, y).multiply(x, y)
power = lambda x, y: _get_np(x, y).power(x, y)
negative = lambda x: _get_np(x).negative(x)
not_equal = lambda x, y: _get_np(x, y).not_equal(x, y)
reciprocal = lambda x: _get_np(x).reciprocal(x)
subtract = lambda x, y: _get_np(x, y).subtract(x, y)
true_divide = lambda x, y: _get_np(x, y).true_divide(x, y)


def divide(x, y):
  np_ = _get_np(x, y)
  dtype = np_.result_type(x.dtype, y.dtype)
  # This is necessary because for integers, divide behaves like true_divide.
  return np_.asarray(np_.divide(x, y), dtype)


def arange(start, stop, step, dtype: tf.DType):
  dtype = _get_dtypes(start, stop, step)[dtype]
  return _get_np(start, stop, step).arange(start, stop, step, dtype=dtype)


def asarray(arr, dtype: tf.DType):
  if _is_poly_dim(arr):
    arr = jax.core.dimension_as_value(arr)
  dtype = _get_dtypes(arr)[dtype]
  return _get_np(arr).asarray(arr, dtype)


def empty(shape, dtype: tf.DType, init: bool):
  del init
  dtype = _get_dtypes(shape)[dtype]
  return _get_np(shape).full(shape, dtype(), dtype=dtype)


def full(shape, fill_value, dtype: tf.DType):
  dtype = _get_dtypes(shape)[dtype]
  return _get_np(shape, fill_value).full(shape, fill_value, dtype=dtype)


def gather(params, indices, axis: int, batch_dims: int):
  """Implement gather in numpy and jax.numpy."""
  if _get_np(params, indices) is np:
    axis = axis if axis >= 0 else params.ndim + axis
    batch_shape = params.shape[:batch_dims]
    take_shape = (
        params.shape[:axis] + indices.shape[batch_dims:] +
        params.shape[axis + 1:])

    if batch_shape:
      res = np.zeros(take_shape, dtype=params.dtype)
      for idx in itertools.product(*[range(v) for v in batch_shape]):
        res[idx] = np.take(params[idx], indices[idx], axis=axis - batch_dims)
    else:
      res = np.take(params, indices, axis=axis)

    return res
  else:
    axis = axis if axis < 0 else -(params.ndim - axis)
    take_fn = lambda p, i: jnp.take(p, i, axis=axis)
    for _ in range(batch_dims):
      take_fn = jax.vmap(take_fn)
    return take_fn(params, indices)


# Reduction ops.
def all_(arr, axis: Union[int, Sequence[int]], keepdims: bool):
  axis = tuple(axis) if isinstance(axis, (list, tuple)) else (axis,)
  return _get_np(arr).all(arr, axis=axis, keepdims=keepdims)


def any_(arr, axis: Union[int, Sequence[int]], keepdims: bool):
  axis = tuple(axis) if isinstance(axis, (list, tuple)) else (axis,)
  return _get_np(arr).any(arr, axis=axis, keepdims=keepdims)


def cumsum(arr, axis: int):
  return _get_np(arr).cumsum(arr, axis=axis)


def max_(arr, axis: Union[int, Sequence[int]], keepdims: bool):
  axis = tuple(axis) if isinstance(axis, (list, tuple)) else (axis,)
  return _get_np(arr).max(arr, axis=axis, keepdims=keepdims)


def min_(arr, axis: Union[int, Sequence[int]], keepdims: bool):
  axis = tuple(axis) if isinstance(axis, (list, tuple)) else (axis,)
  return _get_np(arr).min(arr, axis=axis, keepdims=keepdims)


def prod(arr, axis: Union[int, Sequence[int]], keepdims: bool):
  axis = tuple(axis) if isinstance(axis, (list, tuple)) else (axis,)
  return  _get_np(arr).prod(arr, axis=axis, keepdims=keepdims)


def sum_(arr, axis: Union[int, Sequence[int]], keepdims: bool):
  axis = tuple(axis) if isinstance(axis, (list, tuple)) else (axis,)
  return  _get_np(arr).sum(arr, axis=axis, keepdims=keepdims)


# Array manipulation ops.
def broadcast_to(arr, shape):
  np_ = jnp if any([_is_poly_dim(x) for x in shape]) else _get_np(arr)
  return np_.broadcast_to(arr, shape)

concatenate = lambda arrs, axis: _get_np(*arrs).concatenate(arrs, axis=axis)
expand_dims = lambda arr, axis: _get_np(arr).expand_dims(arr, axis=axis)
flip = lambda arr, axis: _get_np(arr).flip(arr, axis=axis)
roll = lambda arr, shift, axis: _get_np(arr).roll(arr, shift=shift, axis=axis)
split = lambda arr, sections, axis: _get_np(arr).split(arr, sections, axis=axis)
squeeze = lambda arr, axis: _get_np(arr).squeeze(arr, axis=axis)
stack = lambda arrs, axis: _get_np(*arrs).stack(arrs, axis=axis)
tile = lambda arr, reps: _get_np(arr, reps).tile(arr, reps=reps)
where = lambda cond, x, y: _get_np(cond, x, y).where(cond, x, y)


def moveaxis(
    arr,
    source: Union[int, Sequence[int]],
    destination: Union[int, Sequence[int]],
):
  return _get_np(arr).moveaxis(arr, source=source, destination=destination)


def invert_permutation(arr):
  if arr.ndim != 1:
    raise ValueError(f"Expected 1D array, found x.dims={arr.ndim}")
  if _get_np(arr) is np:
    inds = np.zeros_like(arr)
    inds[arr.tolist()] = np.arange(0, arr.shape[0])
    return inds
  else:
    return jnp.zeros_like(arr).at[arr].set(jnp.arange(0, arr.shape[0]))



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Test utils."""

import contextlib

from absl import logging
from absl.testing import parameterized

import jax
import tensorflow as tf


class TestCase(parameterized.TestCase, tf.test.TestCase):
  """Base class for all tests."""

  def setUp(self):
    super().setUp()
    # Ensure that all TF ops are created on the proper device (TPU, GPU or CPU)
    jax_default_device = jax.default_backend().upper()
    tf_logical_devices = tf.config.list_logical_devices(jax_default_device)
    tf_default_device = tf_logical_devices[0]
    logging.info("Running tf2jax converted code on %s.", tf_default_device)
    self.assertEqual(jax_default_device, tf_default_device.device_type)

    with contextlib.ExitStack() as stack:
      stack.enter_context(tf.device(tf_default_device))
      self.addCleanup(stack.pop_all().close)



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tf2jax."""

import itertools

from typing import Tuple

from absl.testing import parameterized

import chex
import jax
import numpy as np

import tensorflow as tf
from tf2jax._src import test_util
from tf2jax._src import tf2jax
import tree


def _parse_version(version: str) -> Tuple[int, ...]:
  return tuple([int(v) for v in version.split(".")])


class OpsTest(test_util.TestCase):

  def _test_convert(self,
                    tf_func,
                    inputs,
                    *,
                    check_shape_only=False,
                    functional=True):
    if not isinstance(inputs, (list, tuple)):
      inputs = (inputs,)

    jax_func, jax_params = tf2jax.convert(
        tf_func, *tree.map_structure(np.zeros_like, inputs))
    if functional:
      self.assertEmpty(jax_params, "Expected no parameters for pure Ops.")

    jax_func = self.variant(jax_func)

    rng = jax.random.PRNGKey(42)
    jax_results, new_jax_params = jax_func(jax_params, *inputs, rng=rng)
    tf_results = tf_func(*inputs)

    # Check outputs
    for tf_res, jax_res in zip(
        tree.flatten(tf_results), tree.flatten(jax_results)):
      self.assertEqual(tf_res.shape, jax_res.shape)
      if not check_shape_only:
        self.assertAllClose(np.asarray(tf_res), jax_res, atol=1e-5)

    return jax_results, new_jax_params

  @chex.variants(with_jit=True, without_jit=True)
  def test_cholesky(self):
    np.random.seed(42)
    inputs = np.stack([
        np.cov(np.random.normal(size=[5, 10]).astype(np.float32))
        for _ in range(3)
    ])

    @tf.function(jit_compile=True)
    def cholesky(x):
      return tf.raw_ops.Cholesky(input=x)

    self._test_convert(cholesky, [inputs])

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("compute_v", True), ("not_compute_v", False)),
          (
              ("unbatched", (5, 5)),
              ("batched", (3, 5, 5)),
              ("more_batched", (2, 3, 5, 5)),
          ),
          named=True,
      ))
  def test_eig(self, compute_v, shape):
    if jax.default_backend().lower() != "cpu":
      self.skipTest("JAX only supports nonsymmetric eigendecomposition on CPU.")

    np.random.seed(42)
    inputs = np.random.normal(size=shape).astype(np.float32)

    # No XLA Eig kernel for CPU.
    is_cpu = jax.default_backend() == "cpu"
    @tf.function(jit_compile=not is_cpu)
    def eig(x):
      return tf.raw_ops.Eig(input=x, compute_v=compute_v, Tout=tf.complex64)

    tf_w, tf_v = eig(tf.constant(inputs))

    jax_eig = tf2jax.convert_functional(eig, np.zeros_like(inputs))
    jax_w, jax_v = self.variant(jax_eig)(inputs)

    # Check eigenvalues are sorted in non-descending order.
    # TODO(b/239787164) Temporarily disabled.
    # self.assertTrue(
    #     np.all((np.absolute(tf_w)[..., :-1] -
    #             np.absolute(tf_w)[..., 1:]) < 1e-5))
    # self.assertTrue(
    #     np.all((np.absolute(jax_w)[..., :-1] -
    #             np.absolute(jax_w)[..., 1:]) < 1e-5))

    # Check eigenvalues so because ordering is not consistent even after sorting
    for idx in itertools.product(*map(range, shape[:-2])):
      w_jax, w_tf = jax_w[idx], tf_w[idx]
      for i in range(shape[-1]):
        self.assertAllClose(min(abs(w_jax - w_tf[i])), 0., atol=1e-5)
        self.assertAllClose(min(abs(w_jax[i] - w_tf)), 0., atol=1e-5)

    if compute_v:
      self.assertAllClose(np.matmul(inputs, tf_v), tf_w[..., None, :] * tf_v)
      self.assertAllClose(np.matmul(inputs, jax_v), jax_w[..., None, :] * jax_v)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("compute_v", True), ("not_compute_v", False)),
          (
              ("unbatched", (5, 5)),
              ("batched", (3, 5, 5)),
              ("more_batched", (2, 3, 5, 5)),
          ),
          named=True,
      ))
  def test_eigh(self, compute_v, shape):
    np.random.seed(42)
    inputs = np.random.normal(size=shape).astype(np.float32)

    @tf.function(jit_compile=True)
    def eigh(x):
      return tf.raw_ops.SelfAdjointEigV2(input=x, compute_v=compute_v)

    tf_w, tf_v = eigh(tf.constant(inputs))

    jax_eigh = tf2jax.convert_functional(eigh, np.zeros_like(inputs))
    jax_w, jax_v = self.variant(jax_eigh)(inputs)

    self.assertTrue(np.all(tf_w[..., :-1] <= tf_w[..., 1:]))
    self.assertTrue(np.all(jax_w[..., :-1] <= jax_w[..., 1:]))

    # Check eigenvalues so because ordering is not consistent even after sorting
    for idx in itertools.product(*map(range, shape[:-2])):
      w_jax, w_tf = jax_w[idx], tf_w[idx]
      for i in range(shape[-1]):
        self.assertAllClose(min(abs(w_jax - w_tf[i])), 0., atol=1e-5)
        self.assertAllClose(min(abs(w_jax[i] - w_tf)), 0., atol=1e-5)

    if compute_v:
      tols = dict(atol=2e-5) if jax.default_backend() in "tpu" else {}
      sym_inputs = np.tril(inputs) + np.swapaxes(np.tril(inputs, -1), -2, -1)
      self.assertAllClose(
          np.matmul(sym_inputs, tf_v), tf_w[..., None, :] * tf_v, **tols)
      self.assertAllClose(
          np.matmul(sym_inputs, jax_v), jax_w[..., None, :] * jax_v, **tols)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (
              ("full_matrices", True),
              ("not_full_matrices", False),
          ),
          (
              ("unbatched", (5, 5)),
              ("batched", (3, 4, 5)),
              ("more_batched", (2, 3, 5, 4)),
          ),
          named=True,
      ))
  def test_qr(self, full_matrices, shape):
    np.random.seed(42)
    inputs = np.random.normal(size=shape).astype(np.float32)

    @tf.function(jit_compile=True)
    def qr(x):
      return tf.raw_ops.Qr(input=x, full_matrices=full_matrices)

    tf_q, tf_r = qr(tf.constant(inputs))

    jax_qr = tf2jax.convert_functional(qr, np.zeros_like(inputs))
    jax_q, jax_r = self.variant(jax_qr)(inputs)

    self.assertEqual(tf_q.shape, jax_q.shape)
    self.assertEqual(tf_r.shape, jax_r.shape)
    self.assertAllClose(jax_q, tf_q)
    self.assertAllClose(jax_r, tf_r)

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (("compute_uv", True), ("not_compute_uv", False)),
          (
              ("full_matrices", True),
              ("not_full_matrices", False),
          ),
          (
              ("unbatched", (3, 3)),
              ("batched", (3, 4, 5)),
              ("more_batched", (2, 3, 5, 4)),
          ),
          named=True,
      ))
  def test_svd(self, compute_uv, full_matrices, shape):
    np.random.seed(42)
    inputs = np.random.normal(size=shape).astype(np.float32)

    @tf.function(jit_compile=True)
    def svd(x):
      return tf.raw_ops.Svd(
          input=x, compute_uv=compute_uv, full_matrices=full_matrices)

    tf_s, tf_u, tf_v = svd(tf.constant(inputs))

    jax_svd = tf2jax.convert_functional(svd, np.zeros_like(inputs))
    jax_s, jax_u, jax_v = self.variant(jax_svd)(inputs)

    self.assertEqual(tf_s.shape, jax_s.shape)
    if compute_uv:
      self.assertEqual(tf_u.shape, jax_u.shape)
      self.assertEqual(tf_v.shape, jax_v.shape)

    # https://numpy.org/doc/stable/reference/generated/numpy.linalg.svd.html)
    def reconstruct(u, s, v):
      with tf.device("CPU"):
        return tf.matmul(
            u[..., :s.shape[-1]] * s[..., None, :],
            v[..., :s.shape[-1]],
            adjoint_b=True)

    # Adapted from tensorflow/python/kernel_tests/linalg/svd_op_test.py
    def compare_singular_vectors(x, y):
      # Singular vectors are only unique up to sign (complex phase factor for
      # complex matrices), so we normalize the sign first.
      sum_of_ratios = np.sum(np.divide(y, x), -2, keepdims=True)
      phases = np.divide(sum_of_ratios, np.abs(sum_of_ratios))
      x = x * phases
      self.assertAllClose(x, y, atol=1e-4)

    tols = dict(atol=1e-5) if jax.default_backend() in ("gpu", "tpu") else {}
    self.assertAllClose(jax_s, tf_s, **tols)
    if compute_uv:
      tf_recon = reconstruct(tf_u, tf_s, tf_v)
      jax_recon = reconstruct(jax_u, jax_s, jax_v)
      self.assertAllClose(tf_recon, jax_recon, atol=1e-5)
      self.assertAllClose(inputs, tf_recon, atol=1e-5)
      self.assertAllClose(inputs, jax_recon, atol=1e-5)

      rank = min(inputs.shape[-2:])
      compare_singular_vectors(tf_u[..., :rank], jax_u[..., :rank])
      compare_singular_vectors(tf_v[..., :rank], jax_v[..., :rank])

  @chex.variants(with_jit=True, without_jit=True)
  @parameterized.named_parameters(
      chex.params_product(
          (
              ("lower", True),
              ("upper", False),
          ),
          (
              ("adjoint", True),
              ("not_adjoint", False),
          ),
          (
              ("unbatched", ((5, 5), (5, 5))),
              ("batched", ((3, 5, 5), (3, 5, 4))),
              ("more_batched", ((2, 3, 5, 5), (2, 3, 5, 6))),
          ),
          named=True,
      ))
  def test_triangular_solve(self, lower, adjoint, shapes):
    lhs_shape, rhs_shape = shapes

    np.random.seed(42)
    inputs = np.random.normal(size=lhs_shape).astype(np.float32)
    rhs_inputs = np.random.normal(size=rhs_shape).astype(np.float32)

    @tf.function(jit_compile=True)
    def triangular_solve(x, rhs):
      return tf.raw_ops.MatrixTriangularSolve(
          matrix=x, rhs=rhs, lower=lower, adjoint=adjoint)

    tf_res = triangular_solve(tf.constant(inputs), tf.constant(rhs_inputs))

    jax_triangular_solve = tf2jax.convert_functional(
        triangular_solve, np.zeros_like(inputs),
        np.zeros_like(rhs_inputs))
    jax_res = self.variant(jax_triangular_solve)(inputs, rhs_inputs)

    self.assertEqual(tf_res.shape, jax_res.shape)
    self.assertAllClose(jax_res, tf_res, atol=1e-5)


if __name__ == "__main__":
  tf.test.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""API of tf2jax."""

from tf2jax._src.config import get_config
from tf2jax._src.config import override_config
from tf2jax._src.config import update_config

from tf2jax._src.tf2jax import AnnotatedFunction

from tf2jax._src.tf2jax import convert
from tf2jax._src.tf2jax import convert_from_restored
from tf2jax._src.tf2jax import convert_functional
from tf2jax._src.tf2jax import convert_functional_from_restored

__version__ = "0.3.2"

#  _________________________________________
# / Please don't use symbols in `_src` they \
# \ are not part of the tf2jax public API.    /
#  -----------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
